#include <SPI.h>  
#include <LiquidCrystal_I2C.h>
#include <Wire.h>
#include <Servo.h>
#include <Encoder.h>

// Game state
enum GameState {
  CODE_GAME,
  LABYRINTH_GAME
};
GameState currentGame = CODE_GAME;

// Pin definitions for the rotary encoder and button
const int clkPin = A11;
const int dtPin = A10;
const int swPin = A12;  // Button press for changing digits

// Configuration for both games' servo
Servo lockServo;
const int servoPin = A8;
int currentAngle = 180;

// LCD Display (shared between games)
LiquidCrystal_I2C lcd(0x27,16,2);

// CODE GAME VARIABLES
Encoder enc(clkPin, dtPin);
long lastPosition = -1;
unsigned long previousMillis = 0;
const long blinkInterval = 500;
bool digitVisible = true;

byte correctNumLEDs[4] = {13,12,11,10};
byte correctPlaceLEDs[4] = {9,8,7,6};

byte code[4] = {0,0,0,0};
byte codeGuess[4] = {0,0,0,0};
byte currentDigit = 0;
byte numGuesses = 0;
boolean correctGuess = false;

// LABYRINTH GAME VARIABLES
const int buttonPins[] = {44,50, 42, 46, 52, 48};
const int numButtons = 6;

int sequences[6][5] = {
  {0, 0, 0, 0, 0},
  {1, 0, 0, 1, 1},
  {1, 0, 0, 1, 0},
  {0, 1, 0, 0, 0},
  {1, 0, 1, 1, 1},
  {0, 1, 1, 1, 1}
};

int currentStep = 0;
int currentSequence;
int buttonArrivalPin;
char startPoint;
boolean labyrinthComplete = false;

void setup() {
  // Initialize shared components
  lcd.init();
  lcd.clear();
  lcd.backlight();
  Serial.begin(9600);
  
  // Initialize servo
  lockServo.attach(servoPin);
  lockServo.write(180);  // Start at 180 degrees (locked)
  currentAngle = 180;    // Track initial angle
  
  // Initialize Code Game components
  for(int i=0; i<=3; i++) {
    pinMode(correctNumLEDs[i], OUTPUT);
    pinMode(correctPlaceLEDs[i], OUTPUT);
  }
  pinMode(swPin, INPUT_PULLUP);
  
  // Initialize Labyrinth Game components
  for (int i = 0; i < numButtons; i++) {
    pinMode(buttonPins[i], INPUT_PULLUP);
  }
  
  startNewGame();
}

void loop() {
  if (currentGame == CODE_GAME) {
    runCodeGame();
  } else {
    runLabyrinthGame();
  }
}

void runCodeGame() {
  long currentPosition = enc.read() / 4;

  if (currentPosition != lastPosition) {
    codeGuess[currentDigit] += (currentPosition > lastPosition) ? 1 : -1;
    if (codeGuess[currentDigit] > 9) codeGuess[currentDigit] = 0;
    if (codeGuess[currentDigit] < 0) codeGuess[currentDigit] = 9;
    
    lastPosition = currentPosition;
    digitVisible = true;
    updatelcdCode();
  }

  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis >= blinkInterval) {
    previousMillis = currentMillis;
    digitVisible = !digitVisible;
    updatelcdCode();
  }

  static bool buttonPressed = false;
  if (digitalRead(swPin) == LOW && !buttonPressed) {
    buttonPressed = true;
    currentDigit = (currentDigit + 1) % 4;
    digitVisible = true;
    if(currentDigit == 0) {
      checkCodeGuess();
      numGuesses++;
    }
    delay(200);
  }

  if (digitalRead(swPin) == HIGH) {
    buttonPressed = false;
  }
  
  if(correctGuess) {
    // Move servo to 135 degrees when code is cracked (smaller movement)
    currentAngle = 135;  
    lockServo.write(currentAngle);
    delay(1000);
    
    // Transition to labyrinth game
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Starting");
    lcd.setCursor(0,1);
    lcd.print("Labyrinth Game!");
    delay(2000);
    
    currentGame = LABYRINTH_GAME;
    setupLabyrinthGame();
  }
}

void runLabyrinthGame() {
  if (currentStep < 5) {
    int buttonState = getButtonState();
    if (buttonState != -1) {
      if (sequences[currentSequence][currentStep] == buttonState) {
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("Correct!");
        
        // Move servo gradually from 135 to 90 degrees (smaller steps)
        currentAngle = 135 - (currentStep + 1) * 9;  
        lockServo.write(currentAngle);
        
        delay(500);
        currentStep++;
        
        // Check if we completed all steps
        if (currentStep == 5) {
          delay(1000);  // Short pause before celebration
          celebrateAndRestart();
          return;
        }
      } else {
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("Erreur!");
        
        // Reset servo to 135 degrees on error
        currentAngle = 135;
        lockServo.write(currentAngle);
        
        delay(1000);
        currentStep = 0;
        currentSequence = random(0, 6);
        setupLabyrinthGame();
      }
    }
  }
}

void celebrateAndRestart() {
  // Victory animation on LCD
  for(int i = 0; i < 3; i++) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Victory!");
    lcd.setCursor(0, 1);
    lcd.print("You did it!");
    delay(500);
    lcd.clear();
    delay(200);
  }
  
  // Flash all LEDs in a pattern
  for(int i = 0; i < 3; i++) {
    // Flash correct number LEDs
    for(int j = 0; j < 4; j++) {
      digitalWrite(correctNumLEDs[j], HIGH);
      delay(100);
    }
    for(int j = 0; j < 4; j++) {
      digitalWrite(correctNumLEDs[j], LOW);
      delay(100);
    }
    
    // Flash correct place LEDs
    for(int j = 0; j < 4; j++) {
      digitalWrite(correctPlaceLEDs[j], HIGH);
      delay(100);
    }
    for(int j = 0; j < 4; j++) {
      digitalWrite(correctPlaceLEDs[j], LOW);
      delay(100);
    }
  }
  
  // Servo victory dance (smaller movements)
  for(int i = 0; i < 5; i++) {  // Increased repetitions
    lockServo.write(90);
    delay(300);
    lockServo.write(95);
    delay(300);
  }
  lockServo.write(90);
  delay(500);
  
  // Show completion message
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Great Job!");
  lcd.setCursor(0, 1);
  lcd.print("Starting Over...");
  delay(2000);
  
  // Reset servo to initial locked position
  currentAngle = 180;
  lockServo.write(currentAngle);
  delay(1000);
  
  // Restart everything
  startNewGame();
}

void startNewGame() {
  // Reset all game variables
  currentGame = CODE_GAME;
  correctGuess = false;
  numGuesses = 0;
  currentDigit = 0;
  lastPosition = enc.read() / 4;
  currentStep = 0;
  labyrinthComplete = false;
  
  // Reset servo to starting position (locked)
  currentAngle = 180;
  lockServo.write(currentAngle);
  
  // Reset display
  lcd.clear();
  
  // Reset code guess
  for(int i=0; i<4; i++) {
    codeGuess[i] = 0;
  }
  
  // Reset LEDs
  updateLEDs(0, 0);
  
  // Show startup animation
  startupAni();
  
  // Generate new code
  generateNewCode();
}

void setupLabyrinthGame() {
  currentSequence = random(0, 6);
  startPoint = (currentSequence < 3) ? 'A' : 'B';
  buttonArrivalPin = buttonPins[random(0, numButtons)];
  
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Depart: ");
  lcd.print(startPoint);
  delay(2000);
  
  displaySequence();
}

void displaySequence() {
  for (int i = 0; i < 5; i++) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print(sequences[currentSequence][i] == 1 ? "Droite -->" : "<-- Gauche");
    delay(1000);
    lcd.clear();
    delay(500);
  }
}

void updatelcdCode() {
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("Code: ");
  
  for (int i=0; i<4; i++) {
    if (i == currentDigit && !digitVisible) {
      lcd.print(" ");
    } else {
      lcd.print(codeGuess[i]);
    }
    if (i < 3) lcd.print(" ");
  }
  
  lcd.setCursor(0,1);
  lcd.print("Digit ");
  lcd.print(currentDigit + 1);
}

void generateNewCode() {
  Serial.print("Code: ");
  for (int i=0; i<=3; i++) {
    code[i] = random(0,10);
    Serial.print(code[i]);
  }
  Serial.println();
  
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("Code to find:");
  lcd.setCursor(0,1);
  for (int i=0; i<4; i++) {
    lcd.print(code[i]);
    if (i < 3) lcd.print(" ");
  }
  delay(3000);
  lcd.clear();
}

void checkCodeGuess() {
  int correctNum = 0;
  int correctPlace = 0;
  int usedDigits[4] = {0,0,0,0};
  
  for (int i=0; i<=3; i++) {
    for (int j=0; j<=3; j++) {
      if (codeGuess[i]==code[j]) {
        if(usedDigits[j]!=1) {
          correctNum++;
          usedDigits[j] = 1;
          break;
        }
      }
    }
  }
  
  for (int i=0; i<=3; i++) {
    if (codeGuess[i]==code[i]) {
      correctPlace++;
      correctNum--;
    }
  }
  
  updateLEDs(correctNum, correctPlace);
  
  if(correctPlace==4) {
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Cracked!");
    correctGuess = true;
  } else {
    correctGuess = false;
  }
}

void updateLEDs(int corNum, int corPla) {
  for(int i=0; i<=3; i++) {
    digitalWrite(correctNumLEDs[i], LOW);
    digitalWrite(correctPlaceLEDs[i], LOW);
  }
  for(int j=0; j<corNum; j++) {
    digitalWrite(correctNumLEDs[j], HIGH);
  }
  for(int k=0; k<corPla; k++) {
    digitalWrite(correctPlaceLEDs[k], HIGH);
  }
}

void startupAni() {
  lcd.setCursor(0,0);
  lcd.print("Code &");
  delay(500);
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("Labyrinth");
  delay(500);
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("Challenge");
  delay(500);
  lcd.clear();
}

int getButtonState() {
  for (int i = 0; i < numButtons; i++) {
    if (digitalRead(buttonPins[i]) == LOW) {
      return i;
    }
  }
  return -1;
}
